using Test
using LibGEOS

@testset "allow_global_context!" begin
    Point(1,2,3)
    LibGEOS.allow_global_context!(false)
    @test_throws ErrorException Point(1,2,3)
    Point(1,2,3, LibGEOS.GEOSContext())
    LibGEOS.allow_global_context!(true) do
        Point(1,2,3)
    end
    @test_throws ErrorException Point(1,2,3)
    LibGEOS.allow_global_context!(true)
    Point(1,2,3)
    Point(1,2,3, LibGEOS.GEOSContext())
    LibGEOS.allow_global_context!(false) do
        @test_throws ErrorException Point(1,2,3)
    end
    Point(1,2,3)
end

@testset "Context mixing" begin
    ctx1 = LibGEOS.GEOSContext()
    ctx2 = LibGEOS.GEOSContext()
    p = [[[-1.,-1],[+1,-1],[+1,+1],[-1,+1],[-1,-1]]]
    p1 = LibGEOS.Polygon(p, ctx1)
    p2 = LibGEOS.Polygon(p, ctx2)

    q = [[[-1.,-1],[+1,-1],[+1,+1],[-1,+1],[-1,-1]]]
    q1 = LibGEOS.Polygon(q, ctx1)
    q2 = LibGEOS.Polygon(q, ctx2)
    @test LibGEOS.intersects(p1, q1)
    @test LibGEOS.intersects(p1, q1, ctx1)
    @test LibGEOS.intersects(p2, q2)
    @test LibGEOS.intersects(p2, q2, ctx2)
    @test_throws ArgumentError LibGEOS.intersects(p1, q2)
    @test_throws ArgumentError LibGEOS.intersects(p2, q1)
end

@testset "show it like you build it" begin
    for geo in [
        readgeom("POINT(0 0)")
        readgeom("MULTIPOINT(0 0, 5 0, 10 0)")
        readgeom("LINESTRING (130 240, 650 240)")
        readgeom("POLYGON EMPTY")
        readgeom("POLYGON ((10 10, 20 40, 90 90, 90 10, 10 10))")
        readgeom("MULTILINESTRING ((5 0, 10 0), (0 0, 5 0))")
        readgeom("GEOMETRYCOLLECTION (LINESTRING (1 2, 2 2), LINESTRING (2 1, 1 1), POLYGON ((0.5 1, 1 2, 1 1, 0.5 1)), POLYGON ((9 2, 9.5 1, 2 1, 2 2, 9 2)))")
        ]
        geo2 = readgeom(sprint(show, geo))
        @test LibGEOS.equals(geo, geo2)
    end
    p = Polygon([[[0.0, 0.0] for _ in 1:10000]])
    buf = IOBuffer()
    print(IOContext(buf, :compact=>true), p)
    seekstart(buf)
    s = read(buf, String)
    @test length(s) < 30
    @test occursin("Polygon(...)", s)
end
